from collections.abc import Generator

from pydantic import BaseModel, RootModel, TypeAdapter
from sseclient import Event

from pharia_skill.csi.inference import (
    ChatEvent,
    ChatParams,
    ChatRequest,
    ChatResponse,
    ChatStreamResponse,
    Completion,
    CompletionAppend,
    CompletionEvent,
    CompletionParams,
    CompletionRequest,
    CompletionStreamResponse,
    ExplanationRequest,
    FinishReason,
    Message,
    MessageAppend,
    MessageBegin,
    TextScore,
    TokenUsage,
)


class DevCompletionStreamResponse(CompletionStreamResponse):
    def __init__(self, stream: Generator[Event, None, None]):
        self._stream = stream

    def next(self) -> CompletionEvent | None:
        if event := next(self._stream, None):
            return completion_event_from_sse(event)
        return None


def completion_event_from_sse(event: Event) -> CompletionEvent:
    match event.event:
        case "append":
            return TypeAdapter(CompletionAppend).validate_json(event.data)
        case "end":
            return FinishReasonDeserializer.model_validate_json(
                event.data
            ).finish_reason
        case "usage":
            return TokenUsageDeserializer.model_validate_json(event.data).usage
        case "error":
            raise ValueError(event.data)
    raise ValueError(f"unknown event type: {event.event}")


class DevChatStreamResponse(ChatStreamResponse):
    def __init__(self, stream: Generator[Event, None, None]):
        self._stream = stream
        super().__init__()

    def next(self) -> ChatEvent | None:
        if event := next(self._stream, None):
            return chat_event_from_sse(event)
        return None


def chat_event_from_sse(event: Event) -> ChatEvent:
    match event.event:
        case "message_begin":
            role = RoleDeserializer.model_validate_json(event.data).role
            return MessageBegin(role)
        case "message_append":
            return TypeAdapter(MessageAppend).validate_json(event.data)
        case "message_end":
            return FinishReasonDeserializer.model_validate_json(
                event.data
            ).finish_reason
        case "usage":
            return TokenUsageDeserializer.model_validate_json(event.data).usage
        case "error":
            raise ValueError(event.data)
    raise ValueError(f"unknown event type: {event.event}")


class FinishReasonDeserializer(BaseModel):
    finish_reason: FinishReason


class TokenUsageDeserializer(BaseModel):
    usage: TokenUsage


class CompletionRequestSerializer(BaseModel):
    model: str
    prompt: str
    params: CompletionParams


class ChatRequestSerializer(BaseModel):
    model: str
    messages: list[Message]
    params: ChatParams


class RoleDeserializer(BaseModel):
    role: str


CompletionRequestListSerializer = RootModel[list[CompletionRequest]]


CompletionListDeserializer = RootModel[list[Completion]]


ChatRequestListSerializer = RootModel[list[ChatRequest]]


ChatListDeserializer = RootModel[list[ChatResponse]]


ExplanationRequestListSerializer = RootModel[list[ExplanationRequest]]


ExplanationListDeserializer = RootModel[list[list[TextScore]]]



from typing import TypeVar, Generic, Union, Optional, Protocol, Tuple, List, Any, Self
from types import TracebackType
from enum import Flag, Enum, auto
from dataclasses import dataclass
from abc import abstractmethod
import weakref

from ..types import Result, Ok, Err, Some


@dataclass
class SelectLanguageRequest:
    """
    Select the detected language for the provided input based on the list of possible languages.
    If no language matches, None is returned.
    
    text: Text input
    languages: All languages that should be considered during detection.
    """
    text: str
    languages: List[str]


def select_language(request: List[SelectLanguageRequest]) -> List[Optional[str]]:
    """
    Select most likely language from a list of supported ISO 639-3language codes.
    
    Afrikaans - "afr",
    Arabic - "ara",
    Azerbaijani - "aze",
    Belarusian - "bel",
    Bengali - "ben",
    Bosnian - "bos",
    Bulgarian - "bul",
    Catalan - "cat",
    Czech - "ces",
    Welsh - "cym",
    Danish - "dan",
    German - "deu",
    Greek - "ell",
    English - "eng",
    Esperanto - "epo",
    Estonian - "est",
    Basque - "eus",
    Persian - "fas",
    Finnish - "fin",
    French - "fra",
    Irish - "gle",
    Gujarati - "guj",
    Hebrew - "heb",
    Hindi - "hin",
    Croatian - "hrv",
    Hungarian - "hun",
    Armenian - "hye",
    Indonesian - "ind",
    Icelandic - "isl",
    Italian - "ita",
    Japanese - "jpn",
    Georgian - "kat",
    Kazakh - "kaz",
    Korean - "kor",
    Latin - "lat",
    Latvian - "lav",
    Lithuanian - "lit",
    Ganda - "lug",
    Marathi - "mar",
    Macedonian - "mkd",
    Mongolian - "mon",
    Maori - "mri",
    Malay - "msa",
    Dutch - "nld",
    Norwegian Nynorsk - "nno",
    Norwegian Bokmål - "nob",
    Punjabi - "pan",
    Polish - "pol",
    Portuguese - "por",
    Romanian - "ron",
    Russian - "rus",
    Slovak - "slk",
    Slovene - "slv",
    Shona - "sna",
    Somali - "som",
    Sotho - "sot",
    Spanish - "spa",
    Serbian - "srp",
    Albanian - "sqi",
    Swahili - "swa",
    Swedish - "swe",
    Tamil - "tam",
    Telugu - "tel",
    Tagalog - "tgl",
    Thai - "tha",
    Tswana - "tsn",
    Tsonga - "tso",
    Turkish - "tur",
    Ukrainian - "ukr",
    Urdu - "urd",
    Vietnamese - "vie",
    Xhosa - "xho",
    Yoruba - "yor",
    Chinese - "zho",
    Zulu - "zul",
    """
    raise NotImplementedError




"""
This module exposes the interfaces for skills to interact with the Pharia Kernel
via the Cognitive System Interface (CSI).
"""

from abc import ABC, abstractmethod
from collections import deque
from collections.abc import Generator
from dataclasses import dataclass, field
from enum import Enum
from types import TracebackType
from typing import Any, Literal, Self

from pydantic import field_validator


@dataclass
class TopLogprobs:
    """Request between 0 and 20 tokens"""

    top: int


NoLogprobs = Literal["no"]
"""Do not return any logprobs"""


SampledLogprobs = Literal["sampled"]
"""Return only the logprob of the tokens which have actually been sampled into the completion."""


Logprobs = TopLogprobs | NoLogprobs | SampledLogprobs
"""Control the logarithmic probabilities you want to have returned."""


@dataclass
class CompletionParams:
    """Completion request parameters.

    Attributes:
        max-tokens (int, optional, default None): The maximum tokens that should be inferred. Note, the backing implementation may return less tokens due to other stop reasons.
        temperature (float, optional, default None): The randomness with which the next token is selected.
        top-k (int, optional, default None): The number of possible next tokens the model will choose from.
        top-p (float, optional, default None): The probability total of next tokens the model will choose from.
        stop (list(str), optional, default []): A list of sequences that, if encountered, the API will stop generating further tokens.
        return_special_tokens (bool, optional, default True): Whether to include special tokens (e.g. <|endoftext|>, <|python_tag|>) in the completion response.
        frequency-penalty (float, optional, default None): The presence penalty reduces the probability of generating tokens that are already present in the generated text respectively prompt. Presence penalty is independent of the number of occurrences. Increase the value to reduce the probability of repeating text.
        presence-penalty (float, optional, default None): The presence penalty reduces the probability of generating tokens that are already present in the generated text respectively prompt. Presence penalty is independent of the number of occurrences. Increase the value to reduce the probability of repeating text.
        logprobs (Logprobs, optional, default NoLogprobs()): Use this to control the logarithmic probabilities you want to have returned. This is useful to figure out how likely it had been that this specific token had been sampled.
    """

    max_tokens: int | None = None
    temperature: float | None = None
    top_k: int | None = None
    top_p: float | None = None
    stop: list[str] = field(default_factory=lambda: list())

    # While the default of this parameters in the api-scheduler is False, we believe that
    # with the introduction of the chat endpoint, the completion endpoint is mostly used for
    # queries where the average user is interested in theses tokens.
    return_special_tokens: bool = True
    frequency_penalty: float | None = None
    presence_penalty: float | None = None
    logprobs: Logprobs = "no"


class FinishReason(str, Enum):
    """The reason the model finished generating.

    Attributes:
        STOP: The model hit a natural stopping point or a provided stop sequence.

        LENGTH: The maximum number of tokens specified in the request was reached.

        CONTENT_FILTER: Content was omitted due to a flag from content filters.
    """

    STOP = "stop"
    LENGTH = "length"
    CONTENT_FILTER = "content_filter"


@dataclass
class Logprob:
    """Logarithmic probability of the token returned in the completion."""

    token: bytes
    logprob: float

    @field_validator("token", mode="before")
    @classmethod
    def convert_token_to_bytes(cls, v: Any) -> Any:
        if isinstance(v, list):
            return bytes(v)
        return v

    def try_as_utf8(self) -> str | None:
        """Try to decode the token as utf-8.

        A token may also represent just a part of an utf-8 character, in which
        case it does not have a valid utf-8 encoding on its own.
        """
        try:
            return self.token.decode("utf-8")
        except UnicodeDecodeError:
            return None


@dataclass
class Distribution:
    sampled: Logprob
    top: list[Logprob]

    @classmethod
    def from_dict(cls, body: dict[str, Any]) -> "Distribution":
        raise NotImplementedError(
            "Not implemented, maybe look into pydantic dataclasses"
        )


@dataclass
class TokenUsage:
    """Usage statistics for the completion request."""

    prompt: int
    completion: int


@dataclass
class CompletionAppend:
    """A chunk of a completion returned by a completion stream.

    Attributes:
        text (str, required): A chunk of the completion text.
        logprobs (list[Distribution], required): Corresponding log probabilities for each token in the completion.
    """

    text: str
    logprobs: list[Distribution]

    @classmethod
    def from_dict(cls, body: dict[str, Any]) -> "CompletionAppend":
        return cls(
            text=body["text"],
            logprobs=body["logprobs"],
        )


@dataclass
class MessageAppend:
    """A chunk of a message generated by the model.

    Attributes:
        content (str, required): A chunk of the message content.
        logprobs (list[Distribution], required): Corresponding log probabilities for each token in the message content.
    """

    content: str
    logprobs: list[Distribution]

    @classmethod
    def from_dict(cls, body: dict[str, Any]) -> "MessageAppend":
        return cls(
            content=body["content"],
            logprobs=body["logprobs"],
        )


CompletionEvent = CompletionAppend | FinishReason | TokenUsage


class CompletionStreamResponse(ABC):
    """Abstract base class for streaming completion responses.

    This class provides the core functionality for streaming completion from a model.
    Concrete implementations only need to implement the `next()` method to provide
    the next event in the stream, and optionally override `__enter__` and `__exit__`
    methods for proper resource management.

    The `__enter__` and `__exit__` methods are particularly important for implementations
    that need to manage external resources. For example, in the `WitCsi` implementation,
    these methods ensure that resources are properly released when the stream is no longer
    needed.
    """

    _finish_reason: FinishReason | None = None
    _usage: TokenUsage | None = None

    def __enter__(self) -> Self:
        """Enter the context manager."""
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        """Exit the context manager and ensure resources are properly cleaned up."""
        pass

    @abstractmethod
    def next(self) -> CompletionEvent | None:
        """Get the next completion event."""
        ...

    def finish_reason(self) -> FinishReason:
        """The reason the model finished generating."""

        if self._finish_reason is None:
            self._consume_stream()
        assert self._finish_reason is not None
        return self._finish_reason

    def usage(self) -> TokenUsage:
        """Usage statistics for the completion request."""

        if self._usage is None:
            self._consume_stream()
        assert self._usage is not None
        return self._usage

    def _consume_stream(self) -> None:
        deque(self.stream(), maxlen=0)
        if self._finish_reason is None or self._usage is None:
            raise ValueError("Invalid event stream")

    def stream(self) -> Generator[CompletionAppend, None, None]:
        """Stream completion chunks."""

        if self._usage:
            raise RuntimeError("The stream has already been consumed")
        while (event := self.next()) is not None:
            match event:
                case CompletionAppend():
                    yield event
                case FinishReason():
                    self._finish_reason = event
                case TokenUsage():
                    self._usage = event
                case _:
                    raise ValueError("Invalid event")


@dataclass
class MessageBegin:
    role: str


ChatEvent = MessageBegin | MessageAppend | FinishReason | TokenUsage


class ChatStreamResponse(ABC):
    """Abstract base class for streaming chat responses.

    This class provides the core functionality for streaming chat from a model.
    Concrete implementations only need to implement the `next()` method to provide
    the next event in the stream, and optionally override `__enter__` and `__exit__`
    methods for proper resource management.

    The `__enter__` and `__exit__` methods are particularly important for implementations
    that need to manage external resources. For example, in the `WitCsi` implementation,
    these methods ensure that resources are properly released when the stream is no longer
    needed.

    The content of the message can be streamed by calling `message_content()`.
    If `finish_reason()` or `usage()` has been called, the stream is consumed.


    Attributes:
        role (str, required): The role of the message.
    """

    role: str

    _finish_reason: FinishReason | None = None
    _usage: TokenUsage | None = None

    def __enter__(self) -> Self:
        """Enter the context manager."""
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        """Exit the context manager and ensure resources are properly cleaned up."""
        pass

    @abstractmethod
    def next(self) -> ChatEvent | None:
        """Get the next chat event."""
        ...

    def __init__(self) -> None:
        first_event = self.next()
        if not isinstance(first_event, MessageBegin):
            raise ValueError("Invalid event stream")
        self.role = first_event.role

    def finish_reason(self) -> FinishReason:
        """The reason the model finished generating."""

        if self._finish_reason is None:
            self._consume_stream()
        assert self._finish_reason is not None
        return self._finish_reason

    def usage(self) -> TokenUsage:
        """Usage statistics for the chat request."""

        if self._usage is None:
            self._consume_stream()
        assert self._usage is not None
        return self._usage

    def _consume_stream(self) -> None:
        deque(self.message_content(), maxlen=0)
        if self._finish_reason is None or self._usage is None:
            raise ValueError("Invalid event stream")

    def message(self) -> Generator[ChatEvent, None, None]:
        """Stream the complete message

        Use this with the @chat decorator to forward all details of the message."""
        yield MessageBegin(self.role)
        yield from self.message_content()
        yield self.finish_reason()
        yield self.usage()

    def message_content(self) -> Generator[MessageAppend, None, None]:
        """Stream the content of the message.

        This does not include the role, the finish reason and usage.
        """
        if self._usage:
            raise RuntimeError("The stream has already been consumed")
        while (event := self.next()) is not None:
            match event:
                case MessageBegin():
                    raise ValueError("Invalid event stream")
                case MessageAppend():
                    yield event
                case FinishReason():
                    self._finish_reason = event
                case TokenUsage():
                    self._usage = event
                    break


@dataclass
class Completion:
    """The result of a completion, including the text generated as well as
    why the model finished completing.

    Attributes:
        text (str, required): The text generated by the model.
        finish-reason (FinishReason, required): The reason the model finished generating.
        logprobs (list[Distribution], required): Contains the logprobs for the sampled and top n tokens, given that `completion-request.params.logprobs` has been set to `sampled` or `top`.
        usage (TokenUsage, required): Usage statistics for the completion request.
    """

    text: str
    finish_reason: FinishReason
    logprobs: list[Distribution]
    usage: TokenUsage

    @classmethod
    def from_dict(cls, body: dict[str, Any]) -> "Completion":
        finish_reason = FinishReason(body["finish_reason"])
        return cls(
            text=body["text"],
            finish_reason=finish_reason,
            logprobs=body["logprobs"],
            usage=body["usage"],
        )


@dataclass
class CompletionRequest:
    """Request a completion from the model

    Attributes:
        model (str, required): Name of model to use.
        prompt (str, required): The text to be completed.
        params (CompletionParams, required): Parameters for the requested completion.
    """

    model: str
    prompt: str
    params: CompletionParams


@dataclass
class ChatParams:
    """Chat request parameters.

    Attributes:
        max-tokens (int, optional, default None):  The maximum tokens that should be inferred. Note, the backing implementation may return less tokens due to other stop reasons.
        temperature (float, optional, default None): The randomness with which the next token is selected.
        top-p (float, optional, default None): The probability total of next tokens the model will choose from.
        frequency-penalty (float, optional, default None): The presence penalty reduces the probability of generating tokens that are already present in the generated text respectively prompt. Presence penalty is independent of the number of occurrences. Increase the value to reduce the probability of repeating text.
        presence-penalty (float, optional, default None): The presence penalty reduces the probability of generating tokens that are already present in the generated text respectively prompt. Presence penalty is independent of the number of occurrences. Increase the value to reduce the probability of repeating text.
        logprobs (Logprobs, optional, default NoLogprobs()): Use this to control the logarithmic probabilities you want to have returned. This is useful to figure out how likely it had been that this specific token had been sampled.
    """

    max_tokens: int | None = None
    temperature: float | None = None
    top_p: float | None = None
    frequency_penalty: float | None = None
    presence_penalty: float | None = None
    logprobs: Logprobs = "no"


class Role(str, Enum):
    """A role used for a message in a chat."""

    User = "user"
    Assistant = "assistant"
    System = "system"


@dataclass
class Message:
    """Describes a message in a chat.

    Parameters:
        role (Role, required): The role of the message.
        content (str, required): The content of the message.
    """

    role: Role
    content: str

    @classmethod
    def user(cls, content: str) -> "Message":
        return cls(role=Role.User, content=content)

    @classmethod
    def assistant(cls, content: str) -> "Message":
        return cls(role=Role.Assistant, content=content)

    @classmethod
    def system(cls, content: str) -> "Message":
        return cls(role=Role.System, content=content)

    @classmethod
    def from_dict(cls, body: dict[str, Any]) -> "Message":
        # the shell csi does not serialize the roles in lowercase
        role = Role(body["role"].lower())
        content = body["content"]
        return cls(role=role, content=content)


@dataclass
class ChatRequest:
    """A chat request that can be sent to the model.

    Attributes:
        model (str, required): Name of model to use.
        messages (list[Message], required): The messages to be sent to the model.
        params (ChatParams, required): Parameters for the requested chat.
    """

    model: str
    messages: list[Message]
    params: ChatParams


@dataclass
class ChatResponse:
    """The result of a chat request.

    Attributes:
        message (Message): The generated message.
        finish_reason (FinishReason): Why the model finished completing.
        logprobs (list[Distribution]): Contains the logprobs for the sampled and top n tokens, given that `chat-request.params.logprobs` has been set to `sampled` or `top`.
        usage (TokenUsage): Usage statistics for the chat request.
    """

    message: Message
    finish_reason: FinishReason
    logprobs: list[Distribution]
    usage: TokenUsage

    @staticmethod
    def from_dict(body: dict[str, Any]) -> "ChatResponse":
        message = Message.from_dict(body["message"])
        finish_reason = FinishReason(body["finish_reason"])
        logprobs = [Distribution.from_dict(logprob) for logprob in body["logprobs"]]
        usage = TokenUsage(body["usage"]["prompt"], body["usage"]["completion"])
        return ChatResponse(message, finish_reason, logprobs, usage)


@dataclass
class TextScore:
    """A range of text with a score indicating how much it influenced the completion.

    Attributes:
        start (int): The start index of the text segment w.r.t. to characters in the prompt.
        length (int): Length of the text segment w.r.t. to characters in the prompt.
        score (float): The score of the text segment, higher means more relevant.
    """

    start: int
    length: int
    score: float


class Granularity(str, Enum):
    """The granularity of the explanation."""

    AUTO = "auto"
    WORD = "word"
    SENTENCE = "sentence"
    PARAGRAPH = "paragraph"


@dataclass
class ExplanationRequest:
    """Request an explanation for the completion.

    Attributes:
        prompt (str): The prompt used for the completion.
        target (str): The completion text.
        model (str): The model used for the completion.
        granularity (Granularity, optional): Controls the length of the ranges which are explained.
    """

    prompt: str
    target: str
    model: str
    granularity: Granularity = Granularity.AUTO



"""
Convert OpenTelemetry spans to the studio format.

The `StudioSpan` model provides the translation from an OpenTelemetry span
to the studio format. It can be set up from an OpenTelemetry span object and
can be (in serialized form) uploaded to the studio collector.

The `StudioSpan` model was copied over from the `ExportedSpan` model in the
intelligence layer. Each occurrence of a field validator marks one translation
step.
"""

import datetime as dt
import json
from collections.abc import Sequence
from enum import Enum
from typing import Any, Self
from uuid import UUID

from opentelemetry.sdk.trace import ReadableSpan
from pydantic import BaseModel, Field, RootModel, field_validator, model_validator


def utc_now() -> dt.datetime:
    """Return datetime object with utc timezone.

    datetime.utcnow() returns a datetime object without timezone, so this function is preferred.
    """
    return dt.datetime.now(dt.timezone.utc)


def double_to_128bit(double_str: str) -> UUID:
    """Convert a 64-bit integer to a 128-bit UUID.

    OpenTelemetry uses 64-bit integers to represent trace IDs and span IDs.
    Studio expects 128-bit UUIDs. The mapping is injective, as the 64-bit integer
    binary representation is doubled to create the 128-bit number.
    """
    double = int(double_str, 16)
    return UUID(int=((double << 64) | double))


class Event(BaseModel):
    name: str
    timestamp: dt.datetime = Field(default_factory=utc_now)

    # Use the attributes field for this as there is no concept of body in OTel events
    body: dict[str, Any] = Field(default_factory=dict, alias="attributes")
    message: str = ""

    @model_validator(mode="after")
    def message_and_name_for_errors(self) -> Self:
        """
        OpenTelemetry does not have the concept of a message for events.

        For errors, set the error message as message and the type as name.
        """
        self.message = self.body.get("exception.message", self.message)
        self.name = self.body.get("exception.type", self.name)
        return self


class TaskSpanAttributes(BaseModel):
    # Needed by the studio to identify task spans
    type: str = "TASK_SPAN"
    input: Any

    # Output is optional as it is not available on exceptions
    output: Any | None = None

    @field_validator("input", mode="before")
    def validate_input(cls, data: str) -> Any:
        """Load a json string into an arbitrary pydantic model.

        OTel attributes do not support dictionary. The input and output is
        serialized to json and is stored as a string.
        """
        return json.loads(data)

    @field_validator("output", mode="before")
    def validate_output(cls, data: str) -> Any:
        return json.loads(data)


class SpanStatus(Enum):
    OK = "OK"
    ERROR = "ERROR"


class Context(BaseModel):
    trace_id: UUID
    span_id: UUID

    @field_validator("trace_id", mode="before")
    def validate_trace_id(cls, data: str) -> UUID:
        """OpenTelemetry uses 128bit hex string to represent trace_id."""
        return UUID(int=int(data, 16))

    @field_validator("span_id", mode="before")
    def validate_span_id(cls, data: str) -> UUID:
        return double_to_128bit(data)


class StudioSpan(BaseModel):
    """Specifies the span/trace data model than can be exported to studio.

    Can be created from an OpenTelemetry span.
    """

    context: Context
    name: str | None
    parent_id: UUID | None
    start_time: dt.datetime
    end_time: dt.datetime
    attributes: TaskSpanAttributes
    events: Sequence[Event]
    status: SpanStatus

    @field_validator("parent_id", mode="before")
    def validate_parent_id(cls, data: str | None) -> UUID | None:
        if data is None:
            return None
        return double_to_128bit(data)

    @field_validator("status", mode="before")
    def validate_status(cls, data: dict[str, str]) -> SpanStatus:
        if data["status_code"] == "OK":
            return SpanStatus.OK
        return SpanStatus.ERROR

    @classmethod
    def from_otel(cls, span: ReadableSpan) -> "StudioSpan":
        """Convert an OpenTelemetry span to the studio format."""
        return cls.model_validate(json.loads(span.to_json()))


StudioSpanList = RootModel[Sequence[StudioSpan]]



"""
Tools integrate LLMs with external services.

For this, three points of interactions are needed, which are represented by three
classes in this module.

1. The model needs to know a about the available tools (`ToolDefinition`).
2. The model needs to be able to call a tool (`ToolCall`).
3. The model needs to know th result of the tool call (`ToolResponse`).
"""

import json
from dataclasses import dataclass
from typing import Any, Sequence

from .response import Response
from .tool import (
    BraveSearch,
    BuiltInTools,
    CodeInterpreter,
    Tool,
    ToolDefinition,
    WolframAlpha,
)


@dataclass
class ToolCall:
    """A tool call as parsed from the response of the model.

    Arguments are not validated against the provided schema.
    """

    name: str

    # tool is put second as for deserialization we always want the dict
    parameters: dict[str, Any] | Tool

    def render(self) -> str:
        """Reconstruct the model response from a parsed tool call.

        There should only be one source of truth. As the response is stored in
        a parsed format, we need to convert it to a prompt string to construct
        the message history for a later interactions with the model.
        """
        if isinstance(self.parameters, dict):
            return json.dumps({"name": self.name, "parameters": self.parameters})
        return self.parameters.render()

    @classmethod
    def from_response(
        cls, text: Response, tools: Sequence[ToolDefinition]
    ) -> "ToolCall | None":
        """Parse a tool call from a message that has been stripped of special tokens.

        While llama3.1 always includes the <|python_tag|> prefix for function calls,
        llama3.3 does not. Therefore, we always try to match a function call from the response,
        even if the python tag is not included. Built in tools are always prefixed with the
        python tag, even by llama3.3.

        This contradicts https://github.com/meta-llama/llama-models/blob/main/models/llama3_3/prompt_format.md#model-response-format-5

        Args:
            text (str): The text of the message stripped of any special tokens.
            python_tag (bool): Whether the message started with the Python Tag.
        """

        if text.python_tag:
            return cls.json_from_text(text.text, tools) or cls.built_in_from_text(
                text.text, tools
            )
        else:
            return cls.json_from_text(text.text, tools)

    @staticmethod
    def built_in_from_text(
        text: str, tools: Sequence[ToolDefinition]
    ) -> "ToolCall | None":
        """Parse a tool call from a message that started with the Python Tag."""
        if BraveSearch in tools and (brave_search := BraveSearch.try_from_text(text)):
            return ToolCall("brave_search", brave_search)
        elif WolframAlpha in tools and (wolfram := WolframAlpha.try_from_text(text)):
            return ToolCall("wolfram_alpha", wolfram)
        elif CodeInterpreter in tools:
            return ToolCall("code_interpreter", CodeInterpreter(src=text.strip()))
        return None

    @classmethod
    def json_from_text(
        cls, response: str, tools: Sequence[ToolDefinition]
    ) -> "ToolCall | None":
        """Try parsing a tool call into one of the user provided tools.

        Raise a pydantic validation error if the model tries a call that
        can not be parsed into the provided schema.
        """
        try:
            data = json.loads(response)
            name = data["name"]
            parameters = data["parameters"]
            tool = next(
                (t for t in tools if t.name() == name and t not in BuiltInTools), None
            )
            if tool:
                if isinstance(tool, type):
                    parameters = tool(**parameters)

                return ToolCall(name, parameters)
        except (json.JSONDecodeError, KeyError):
            pass

        return None



from .chat_decorator import chat
from .message_stream_decorator import message_stream
from .writer import (
    MessageAppend,
    MessageBegin,
    MessageEnd,
    MessageItem,
    MessageWriter,
)

__all__ = [
    "chat",
    "message_stream",
    "MessageBegin",
    "MessageAppend",
    "MessageEnd",
    "MessageWriter",
    "MessageItem",
]



"""
This module exposes the Python version of the Cognitive System Interface (CSI) and all related types,
which Skills use to interact with the Pharia Kernel.

While developers only interact with these types and the CSI protocol, the SDK injects the `wit_csi.WitCsi`
implementation at runtime, which does a translation step from the Python types we offer in the SDK
to those in `wit.imports`, which are automatically generated from the WIT world via `componentize-py`.
"""

from .chunking import Chunk, ChunkParams, ChunkRequest
from .csi import Csi
from .document_index import (
    After,
    AtOrAfter,
    AtOrBefore,
    Before,
    Cursor,
    Document,
    DocumentPath,
    EqualTo,
    FilterCondition,
    GreaterThan,
    GreaterThanOrEqualTo,
    Image,
    IndexPath,
    IsNull,
    JsonSerializable,
    LessThan,
    LessThanOrEqualTo,
    MetadataFilter,
    Modality,
    SearchFilter,
    SearchRequest,
    SearchResult,
    Text,
    With,
    WithOneOf,
    Without,
)
from .inference import (
    ChatEvent,
    ChatParams,
    ChatRequest,
    ChatResponse,
    Completion,
    CompletionParams,
    CompletionRequest,
    Distribution,
    ExplanationRequest,
    FinishReason,
    Granularity,
    Logprob,
    Logprobs,
    Message,
    NoLogprobs,
    Role,
    SampledLogprobs,
    TextScore,
    TokenUsage,
    TopLogprobs,
)
from .language import Language, SelectLanguageRequest

__all__ = [
    "After",
    "AtOrAfter",
    "AtOrBefore",
    "Before",
    "ChatEvent",
    "ChatParams",
    "ChatRequest",
    "ChatResponse",
    "Chunk",
    "ChunkParams",
    "ChunkRequest",
    "Completion",
    "CompletionParams",
    "CompletionRequest",
    "Csi",
    "Cursor",
    "Distribution",
    "Document",
    "DocumentPath",
    "EqualTo",
    "ExplanationRequest",
    "FinishReason",
    "Granularity",
    "GreaterThan",
    "GreaterThanOrEqualTo",
    "Image",
    "IndexPath",
    "IsNull",
    "JsonSerializable",
    "Language",
    "LessThan",
    "LessThanOrEqualTo",
    "Logprob",
    "Logprobs",
    "Message",
    "MetadataFilter",
    "FilterCondition",
    "Modality",
    "NoLogprobs",
    "Role",
    "SampledLogprobs",
    "SearchFilter",
    "SearchRequest",
    "SearchResult",
    "SelectLanguageRequest",
    "Text",
    "TextScore",
    "TokenUsage",
    "TopLogprobs",
    "Without",
    "With",
    "WithOneOf",
]



from typing import TypeVar, Generic, Union, Optional, Protocol, Tuple, List, Any, Self
from types import TracebackType
from enum import Flag, Enum, auto
from dataclasses import dataclass
from abc import abstractmethod
import weakref

from ..types import Result, Ok, Err, Some


@dataclass
class TextScore:
    """
    A score for a text segment.
    """
    start: int
    length: int
    score: float

class Granularity(Enum):
    """
    At which granularity should the target be explained in terms of the prompt.
    If you choose, for example, [`granularity.sentence`] then we report the importance score of each
    sentence in the prompt towards generating the target output.
    The default is [`granularity.auto`] which means we will try to find the granularity that
    brings you closest to around 30 explanations. For large prompts, this would likely
    be sentences. For short prompts this might be individual words or even tokens.
    """
    AUTO = 0
    WORD = 1
    SENTENCE = 2
    PARAGRAPH = 3

@dataclass
class ExplanationRequest:
    prompt: str
    target: str
    model: str
    granularity: Granularity

class FinishReason(Enum):
    """
    The reason the model finished generating
    """
    STOP = 0
    LENGTH = 1
    CONTENT_FILTER = 2

@dataclass
class Logprob:
    token: bytes
    logprob: float

@dataclass
class Distribution:
    sampled: Logprob
    top: List[Logprob]

@dataclass
class TokenUsage:
    prompt: int
    completion: int

@dataclass
class Completion:
    """
    The result of a completion, including the text generated as well as
    why the model finished completing.
    """
    text: str
    finish_reason: FinishReason
    logprobs: List[Distribution]
    usage: TokenUsage


@dataclass
class Logprobs_No:
    pass


@dataclass
class Logprobs_Sampled:
    pass


@dataclass
class Logprobs_Top:
    value: int


Logprobs = Union[Logprobs_No, Logprobs_Sampled, Logprobs_Top]


@dataclass
class CompletionParams:
    """
    Completion request parameters
    """
    max_tokens: Optional[int]
    temperature: Optional[float]
    top_k: Optional[int]
    top_p: Optional[float]
    stop: List[str]
    return_special_tokens: bool
    frequency_penalty: Optional[float]
    presence_penalty: Optional[float]
    logprobs: Logprobs

@dataclass
class CompletionRequest:
    """
    Completion request parameters
    """
    model: str
    prompt: str
    params: CompletionParams

@dataclass
class CompletionAppend:
    """
    A chunk of a completion returned by a completion stream.
    """
    text: str
    logprobs: List[Distribution]


@dataclass
class CompletionEvent_Append:
    value: CompletionAppend


@dataclass
class CompletionEvent_End:
    value: FinishReason


@dataclass
class CompletionEvent_Usage:
    value: TokenUsage


CompletionEvent = Union[CompletionEvent_Append, CompletionEvent_End, CompletionEvent_Usage]
"""
An event emitted by a completion stream.
"""


class CompletionStream:
    """
    Allows for streaming completion tokens as they are generated.
    """
    
    def __init__(self, init: CompletionRequest) -> None:
        """
        Creates a new completion-stream resource for a given completion-request.
        """
        raise NotImplementedError

    def next(self) -> Optional[CompletionEvent]:
        """
        Returns the next completion-event from the completion-stream.
        Will return None if the stream has finished.
        """
        raise NotImplementedError
    def __enter__(self) -> Self:
        """Returns self"""
        return self
                                
    def __exit__(self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None) -> bool | None:
        """
        Release this resource.
        """
        raise NotImplementedError


@dataclass
class Message:
    role: str
    content: str

@dataclass
class ChatParams:
    max_tokens: Optional[int]
    temperature: Optional[float]
    top_p: Optional[float]
    frequency_penalty: Optional[float]
    presence_penalty: Optional[float]
    logprobs: Logprobs

@dataclass
class ChatResponse:
    """
    The result of a chat response, including the message generated as well as
    why the model finished completing.
    """
    message: Message
    finish_reason: FinishReason
    logprobs: List[Distribution]
    usage: TokenUsage

@dataclass
class ChatRequest:
    model: str
    messages: List[Message]
    params: ChatParams

@dataclass
class MessageAppend:
    """
    A chunk of a message generated by the model.
    """
    content: str
    logprobs: List[Distribution]


@dataclass
class ChatEvent_MessageBegin:
    value: str


@dataclass
class ChatEvent_MessageAppend:
    value: MessageAppend


@dataclass
class ChatEvent_MessageEnd:
    value: FinishReason


@dataclass
class ChatEvent_Usage:
    value: TokenUsage


ChatEvent = Union[ChatEvent_MessageBegin, ChatEvent_MessageAppend, ChatEvent_MessageEnd, ChatEvent_Usage]
"""
An event emitted by the chat-stream resource.
"""


class ChatStream:
    
    def __init__(self, init: ChatRequest) -> None:
        """
        Creates a new chat-stream resource for a given chat-request.
        """
        raise NotImplementedError

    def next(self) -> Optional[ChatEvent]:
        """
        Returns the next chat-event from the chat-stream.
        Will return None if the stream has finished.
        """
        raise NotImplementedError
    def __enter__(self) -> Self:
        """Returns self"""
        return self
                                
    def __exit__(self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None) -> bool | None:
        """
        Release this resource.
        """
        raise NotImplementedError



def explain(request: List[ExplanationRequest]) -> List[List[TextScore]]:
    """
    Better understand the source of a completion, specifically on how much each section of a prompt impacts each token of the completion.
    """
    raise NotImplementedError

def complete(requests: List[CompletionRequest]) -> List[Completion]:
    raise NotImplementedError

def chat(requests: List[ChatRequest]) -> List[ChatResponse]:
    raise NotImplementedError




from dataclasses import dataclass
from enum import Enum


class Language(str, Enum):
    """ISO 639-3 language."""

    Afrikaans = "afr"
    Arabic = "ara"
    Azerbaijani = "aze"
    Belarusian = "bel"
    Bengali = "ben"
    Bosnian = "bos"
    Bulgarian = "bul"
    Catalan = "cat"
    Czech = "ces"
    Welsh = "cym"
    Danish = "dan"
    German = "deu"
    Greek = "ell"
    English = "eng"
    Esperanto = "epo"
    Estonian = "est"
    Basque = "eus"
    Persian = "fas"
    Finnish = "fin"
    French = "fra"
    Irish = "gle"
    Gujarati = "guj"
    Hebrew = "heb"
    Hindi = "hin"
    Croatian = "hrv"
    Hungarian = "hun"
    Armenian = "hye"
    Indonesian = "ind"
    Icelandic = "isl"
    Italian = "ita"
    Japanese = "jpn"
    Georgian = "kat"
    Kazakh = "kaz"
    Korean = "kor"
    Latin = "lat"
    Latvian = "lav"
    Lithuanian = "lit"
    Ganda = "lug"
    Marathi = "mar"
    Macedonian = "mkd"
    Mongolian = "mon"
    Maori = "mri"
    Malay = "msa"
    Dutch = "nld"
    NorwegianNynorsk = "nno"
    NorwegianBokmål = "nob"
    Punjabi = "pan"
    Polish = "pol"
    Portuguese = "por"
    Romanian = "ron"
    Russian = "rus"
    Slovak = "slk"
    Slovene = "slv"
    Shona = "sna"
    Somali = "som"
    Sotho = "sot"
    Spanish = "spa"
    Serbian = "srp"
    Albanian = "sqi"
    Swahili = "swa"
    Swedish = "swe"
    Tamil = "tam"
    Telugu = "tel"
    Tagalog = "tgl"
    Thai = "tha"
    Tswana = "tsn"
    Tsonga = "tso"
    Turkish = "tur"
    Ukrainian = "ukr"
    Urdu = "urd"
    Vietnamese = "vie"
    Xhosa = "xho"
    Yoruba = "yor"
    Chinese = "zho"
    Zulu = "zul"


@dataclass
class SelectLanguageRequest:
    """
    Select the detected language for the provided input based on the list of possible languages.
    If no language matches, None is returned.

    Attributes:
        text (str, required): Text input
        languages (list[Language], required): All languages that should be considered during detection.
    """

    text: str
    languages: list[Language]



import inspect
from types import GeneratorType
from typing import Any, Callable, Generator, TypeVar

from pydantic import BaseModel

from pharia_skill import Csi
from pharia_skill.csi.inference import (
    ChatEvent,
    FinishReason,
    MessageAppend,
    MessageBegin,
    TokenUsage,
)
from pharia_skill.stream.message_stream_decorator import message_stream
from pharia_skill.stream.writer import MessageWriter

Input = TypeVar("Input", bound=BaseModel)
Output = TypeVar("Output", bound=BaseModel)


def chat(
    func: Callable[[Csi, Input], Generator[ChatEvent, None, Output | None]],
) -> Callable[[Csi, Input], Generator[ChatEvent, None, Output | None]]:
    """Turn a generator into a Skill that streams it outputs.

    This decorator offers a different interface than @message_stream.
    While @message_stream injects a `MessageWriter` that you can use to write
    intermediate results to, this decorator lets you write a generator that yields
    intermediate results and optionally returns a result.

    The decorated function must be a typed generator function. It must have exactly two arguments.
    The first argument must be of type `Csi` and the second argument must be a Pydantic model.
    The items type of the generator must be a `ChatEvent`. The return type must be a Pydantic model or `None`.

    Example::

        @chat
        def haiku_stream(csi: Csi, input: HaikuInput) -> Generator[ChatEvent, None, HaikuOutput]:
            model = "llama-3.1-8b-instruct"
            messages = [
                Message.system("You are a poet who strictly speaks in haikus."),
                Message.user(input.topic),
            ]
            params = ChatParams()
            with csi.chat_stream(model, messages, params) as response:
                yield from response.message()

            return HaikuOutput(anything="anything")
    """
    decorated = message_stream(to_message_stream(func))

    assert "MessageStream" not in func.__globals__, (
        "Make sure to decorate with either `@chat` or `@message_stream` once."
    )

    func.__globals__["MessageStream"] = decorated.__globals__["MessageStream"]
    del decorated.__globals__["MessageStream"]

    return func


def to_message_stream(
    func: Callable[[Csi, Input], Generator[ChatEvent, None, Output | None]],
) -> Callable[[Csi, MessageWriter[Output], Input], None]:
    """Convert a chat skill a user would write to a function that can be decorated with @message_stream.

    Having this separated from the decorator allows for easier testing of the generated function.
    """

    signature = list(inspect.signature(func).parameters.values())
    assert len(signature) == 2, "Chat Skills must have exactly two arguments."

    input_model = signature[1].annotation
    assert isinstance(input_model, type) and issubclass(input_model, BaseModel), (
        "The second argument must be a Pydantic model, found: " + str(input_model)
    )

    def inner(csi: Csi, writer: MessageWriter[Output], input: input_model) -> None:  # type: ignore
        try:
            wrapper = GeneratorWrapper(func(csi, input))
        except ValueError:
            raise ValueError(
                "A Skill decorated with @chat must be a generator. ",
                "Make sure to yield from your Skill, otherwise you should use the @skill decorator.",
            )
        for event in wrapper:
            match event:
                case MessageBegin(role):
                    writer.begin_message(role)
                case MessageAppend(content):
                    writer.append_to_message(content)
                case FinishReason() | TokenUsage():
                    # This decorator is opinionated: If you use it, you "simply" want to stream a single completion.
                    # The caller does not get to see the token usage or the finish reason.
                    pass
                case _:
                    raise ValueError(
                        f"Yielded unexpected event of type {type(event)} in the @chat decorator. "
                        f"The @chat decorator only supports MessageBegin, MessageAppend, FinishReason, and TokenUsage events, got: {event}"
                    )

        # Write the return message of the function
        writer.end_message(wrapper.value)

    return inner


class GeneratorWrapper:
    """Helps to extract the return value of a generator.

    See https://peps.python.org/pep-0380/#use-of-stopiteration-to-return-values
    """

    def __init__(self, gen: Generator[Any, None, Any]):
        if not isinstance(gen, GeneratorType):
            raise ValueError
        self.gen = gen

    def __iter__(self) -> Any:
        self.value = yield from self.gen
        return self.value



from dataclasses import dataclass
from typing import Generic, Protocol, TypeVar

from pydantic import BaseModel

Payload = TypeVar("Payload", bound=BaseModel | None)


@dataclass
class MessageBegin:
    role: str | None


@dataclass
class MessageAppend:
    text: str


@dataclass
class MessageEnd(Generic[Payload]):
    payload: Payload | None


MessageItem = MessageBegin | MessageAppend | MessageEnd[Payload]


class MessageWriter(Protocol, Generic[Payload]):
    """Write messages to the output stream."""

    def write(self, item: MessageItem[Payload]) -> None: ...

    def begin_message(self, role: str | None = None) -> None:
        self.write(MessageBegin(role))

    def append_to_message(self, text: str) -> None:
        self.write(MessageAppend(text))

    def end_message(self, payload: Payload | None = None) -> None:
        self.write(MessageEnd(payload))



from typing import TypeVar, Generic, Union, Optional, Protocol, Tuple, List, Any, Self
from types import TracebackType
from enum import Flag, Enum, auto
from dataclasses import dataclass
from abc import abstractmethod
import weakref

from ..types import Result, Ok, Err, Some


@dataclass
class ChunkParams:
    """
    Chunking parameters
    """
    model: str
    max_tokens: int
    overlap: int

@dataclass
class ChunkRequest:
    text: str
    params: ChunkParams

@dataclass
class ChunkWithOffsetRequest:
    text: str
    params: ChunkParams
    character_offsets: bool

@dataclass
class ChunkWithOffset:
    text: str
    byte_offset: int
    character_offset: Optional[int]


def chunk(request: List[ChunkRequest]) -> List[List[str]]:
    raise NotImplementedError

def chunk_with_offsets(request: List[ChunkWithOffsetRequest]) -> List[List[ChunkWithOffset]]:
    raise NotImplementedError




from collections.abc import Sequence
from typing import Protocol

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import (
    SimpleSpanProcessor,
    SpanExporter,
    SpanExportResult,
)

from pharia_skill.studio.span import StudioSpan


class SpanClient(Protocol):
    """Client that can submit spans.

    Separating the collection of spans from the uploading allows
    for better modularity and testability.
    """

    def submit_spans(self, spans: Sequence[StudioSpan]) -> None: ...


class StudioExporter(SpanExporter):
    """An OpenTelemetry exporter that uploads spans to Studio.

    The exporter will create a project on setup if it does not exist yet.
    It is generic over the client, allowing to decouple the collection
    from the uploading step.
    """

    def __init__(self, client: SpanClient):
        self.spans: dict[int, list[ReadableSpan]] = {}
        self.client = client

    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        """Store spans in the exporter and upload them to Studio when the exporter shuts down.

        Studio is complaining about duplicate IDs when uploading traces with the same `span_id`
        in separate requests. Therefore, we store the spans and only flush when the root span ends.

        Args:
            spans (Sequence[ReadableSpan], required): stores a sequence of readable spans for upload
        """
        for span in spans:
            if span.context is None:
                raise ValueError("Span has no context")
            self._store_span(span)
            if span.parent is None:
                self._flush_trace(span.context.trace_id)

        return SpanExportResult.SUCCESS

    def _store_span(self, span: ReadableSpan) -> None:
        """Spans are grouped by trace_id for storage."""
        if span.context is None:
            raise ValueError("Span has no context")

        if (trace_id := span.context.trace_id) not in self.spans:
            self.spans[trace_id] = []
        self.spans[trace_id].append(span)

    def _flush_trace(self, trace_id: int) -> None:
        spans = self.spans.pop(trace_id)
        studio_spans = [StudioSpan.from_otel(span) for span in spans]
        self.client.submit_spans(studio_spans)

    def shutdown(self) -> None:
        """Will be called at the end of a session.

        There must not be any open spans left, all open spans should have been called with a parent,
        which has triggered the upload already.
        """
        assert len(self.spans) == 0, "No spans should be left in the exporter"
        self.spans.clear()


class StudioSpanProcessor(SimpleSpanProcessor):
    """Signal that a processor has been registered by the SDK."""

    pass



from dataclasses import asdict, dataclass
from typing import Any

from pydantic import Field, model_serializer


@dataclass
class ChunkParams:
    """Chunking parameters.

    Attributes:
        model (str, required): The name of the model the chunk is intended to be used for. This must be a known model.
        max_tokens (int, required): The maximum number of tokens that should be returned per chunk.
        overlap (int, optional, default 0): The amount of allowed overlap between chunks. Must be less than max_tokens. By default, there is no overlap between chunks.
    """

    model: str
    max_tokens: int
    overlap: int = 0


@dataclass
class ChunkRequest:
    """Chunking request parameters.

    Attributes:
        text (str, required): The text to be chunked.
        params (ChunkParams, required): Parameter used for chunking.
    """

    text: str
    params: ChunkParams

    @model_serializer()
    def serialize(self) -> dict[str, Any]:
        return {
            "text": self.text,
            "params": asdict(self.params),
            "character_offsets": True,
        }


@dataclass
class Chunk:
    """Chunk object with offset information.

    Attributes:
        text (str, required): The text that was chunked
        offset (int, required): The character offset relative to the start of the original text
    """

    text: str
    offset: int = Field(validation_alias="character_offset")



from typing import TypeVar, Generic, Union, Optional, Protocol, Tuple, List, Any, Self
from types import TracebackType
from enum import Flag, Enum, auto
from dataclasses import dataclass
from abc import abstractmethod
import weakref

from .types import Result, Ok, Err, Some



class SkillAll(Protocol):
    pass



"""
This module encapsulates knowledge about the structure of llama3 prompts and responses.
"""

from dataclasses import dataclass
from enum import Enum


class SpecialTokens(str, Enum):
    EndOfTurn = "<|eot_id|>"
    EndOfMessage = "<|eom_id|>"
    PythonTag = "<|python_tag|>"
    BeginOfText = "<|begin_of_text|>"
    StartHeader = "<|start_header_id|>"
    EndHeader = "<|end_header_id|>"


RawResponse = str
"""Unparsed response as received from the model.

Contains special tokens and whitespace.
"""


@dataclass(frozen=True)
class Response:
    """Inner part of the completion.

    The text is stripped from all special tokens and stop reasons.
    Information about the presence of the Python tag is stored in a separate attribute.
    """

    text: str
    python_tag: bool

    @staticmethod
    def from_raw(raw: RawResponse) -> "Response":
        """Parse a raw response (as received from the model)."""
        raw = raw.replace(SpecialTokens.EndOfTurn, "")
        raw = raw.replace(SpecialTokens.EndOfMessage, "")

        # Strip all whitespace from the message.
        # We assume that no model response or tool call should
        # start or end with whitespace.
        raw = raw.strip()
        python_tag = raw.startswith(SpecialTokens.PythonTag)
        text = raw.replace(SpecialTokens.PythonTag, "")
        return Response(text, python_tag)



from .message import AssistantMessage, Role, ToolMessage, UserMessage
from .request import ChatRequest, ChatResponse
from .response import SpecialTokens
from .tool import (
    BraveSearch,
    CodeInterpreter,
    JsonSchema,
    Tool,
    ToolDefinition,
    WolframAlpha,
)
from .tool_call import ToolCall

__all__ = [
    "ChatRequest",
    "ChatResponse",
    "UserMessage",
    "Role",
    "AssistantMessage",
    "ToolCall",
    "ToolDefinition",
    "ToolMessage",
    "BraveSearch",
    "SpecialTokens",
    "JsonSchema",
    "CodeInterpreter",
    "WolframAlpha",
    "Tool",
]
